import copy
import torch
import numpy as np

from torch import nn
from torch.optim.lr_scheduler import ExponentialLR

"""DEVICE USED IN TRAINING"""
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

"""Custom made functions (scheduler_to, optimizer to). With the purpose to put tensors on a selected device"""
"""Pytorch --> Feature request --> Open: https://github.com/pytorch/pytorch/issues/8741"""


def scheduler_to(sched, device):
    for param in sched.__dict__.values():
        if isinstance(param, torch.Tensor):
            param.data = param.data.to(device)
            if param._grad is not None:
                param._grad.data = param._grad.data.to(device)


def optimizer_to(optim, device):
    for param in optim.state.values():
        # Not sure there are any global tensors in the state dict
        if isinstance(param, torch.Tensor):
            param.data = param.data.to(device)
            if param._grad is not None:
                param._grad.data = param._grad.data.to(device)
        elif isinstance(param, dict):
            for subparam in param.values():
                if isinstance(subparam, torch.Tensor):
                    subparam.data = subparam.data.to(device)
                    if subparam._grad is not None:
                        subparam._grad.data = subparam._grad.data.to(device)


def create_dataset(df):
    sequences = df.astype(np.float32).tolist()
    dataset = [torch.tensor(s).unsqueeze(1).float() for s in sequences]
    n_seq, seq_len, n_features = torch.stack(dataset).shape
    return dataset, seq_len, n_features


def train_model(model, optimizer, train_dataset, n_epochs):
    scheduler = ExponentialLR(optimizer, gamma=0.9)
    scheduler_to(scheduler, device)
    criterion = nn.L1Loss(reduction='sum').to(device)
    history = dict(train=[], val=[])
    best_model_wts = copy.deepcopy(model.state_dict())
    best_loss = 100000.0

    for epoch in range(1, n_epochs + 1):

        """Training the model"""
        model = model.train()
        train_losses = []

        for seq_true in train_dataset:
            optimizer.zero_grad()
            seq_true = seq_true.to(device)
            seq_pred = model(seq_true)
            # https://pytorch.org/tutorials/beginner/former_torchies/autograd_tutorial.html
            seq_pred.requires_grad_(True)
            seq_true.requires_grad_(True)

            loss = criterion(seq_pred, seq_true)

            loss.backward()
            optimizer.step()
            train_losses.append(loss.item())

        scheduler.step()
        train_loss = np.mean(train_losses)
        history['train'].append(train_loss)

        if train_loss < best_loss:
            best_loss = train_loss
            best_model_wts = copy.deepcopy(model.state_dict())

    model.load_state_dict(best_model_wts)
    return model.eval(), history


def predict(model, dataset):
    predictions, losses, original = [], [], []
    criterion = nn.L1Loss(reduction='sum').to(device)
    with torch.no_grad():
        model = model.eval()
        for seq_true in dataset:
            seq_true = seq_true.to(device)
            seq_pred = model(seq_true)
            loss = criterion(seq_pred, seq_true)
            predictions.append(seq_pred.cpu().numpy().flatten())
            losses.append(loss.item())
            original.append(seq_true.cpu().numpy().flatten())
    return predictions, losses, original



import math
import sys
import torch
import nianet.helper as helper

from niapy import Runner
from niapy.problems import Problem
from niapy.algorithms.basic import *
from niapy.algorithms.modified import *
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error
from nianet.autoencoder import Autoencoder


class AutoencoderArchitecture(Problem):

    def __init__(self, dimension, X_train, y_train, X_test, y_test, alpha=0.99):
        super().__init__(dimension=dimension, lower=0, upper=1)
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        self.alpha = alpha
        self.iteration = 0

    def _evaluate(self, solution):
        print("=================================================================================================")
        print(f"ITERATION IS: {self.iteration}")
        self.iteration += 1

        X_train_sequence, seq_len, n_features = helper.create_dataset(self.X_train)
        X_test_sequence, _, _ = helper.create_dataset(self.X_test)

        dataset_shape = self.X_train.shape
        model = Autoencoder(solution, dataset_shape)

        """Punishing bad decisions"""
        if len(model.encoding_layers) == 0 or len(model.decoding_layers) == 0:
            fitness = sys.maxsize
            print(f"Fitness: {fitness}")
            return fitness

        model = model.to(helper.device)

        torch.cuda.empty_cache()
        helper.optimizer_to(model.optimizer, helper.device)

        model, history = helper.train_model(
            model,
            model.optimizer,
            X_train_sequence,
            n_epochs=model.epochs
        )

        # Known problem: https://discuss.pytorch.org/t/why-my-model-returns-nan/24329/5
        if math.isnan(min(history['train'])):
            fitness = sys.maxsize
            print(f"Fitness: {fitness}")
            return fitness

        else:
            predictions, losses, original = helper.predict(model, X_test_sequence)
            RMSE = mean_squared_error(original, predictions, squared=False)
            complexity = (model.epochs ** 2) + (model.layers * 100) + (model.bottleneck_size * 10)
            fitness = (RMSE * 1000) + (complexity / 100)
            print(f"RMSE: {RMSE}")
            print(f"Complexity: {complexity}")
            print(f"Fitness: {fitness}")

            return fitness


def find_architecture(df_dataset):
    target = df_dataset.loc[:, df_dataset.columns == 'target']
    data = df_dataset.loc[:, df_dataset.columns != 'target']
    """
    Dimensionality:
    y1: topology shape,
    y2: number of neurons per layer
    y3: number of layers,
    y4: activation function
    y5: number of epochs,
    y6: learning rate
    y7: optimizer algorithm.
    """
    DIMENSIONALITY = 7
    data = StandardScaler().fit_transform(data)
    X_train, X_test, y_train, y_test = train_test_split(data, target, test_size=0.33, random_state=1234)

    runner = Runner(
        dimension=DIMENSIONALITY,
        max_evals=100,
        runs=2,
        algorithms=[
            ParticleSwarmAlgorithm(),
            DifferentialEvolution(),
            FireflyAlgorithm(),
            SelfAdaptiveDifferentialEvolution(),
            GeneticAlgorithm()
        ],
        problems=[
            AutoencoderArchitecture(DIMENSIONALITY, X_train, y_train, X_test, y_test)
        ]
    )

    print("=================================================================================================")
    final_solutions = runner.run(export='json', verbose=True)
    best_fitness = sys.maxsize
    best_solution = None

    for algorithm in final_solutions:
        fitness = final_solutions[algorithm]['AutoencoderArchitecture'][0][1]
        solution = final_solutions[algorithm]['AutoencoderArchitecture'][0][0]
        print(f"{algorithm}'s fitness: {fitness}")
        print(f"{algorithm}'s solution: {solution}")

        if best_fitness > fitness:
            best_fitness = fitness
            best_solution = final_solutions[algorithm]['AutoencoderArchitecture'][0][0]

    print("=================================================================================================")
    model = Autoencoder(best_solution, X_train.shape)
    print(f"Best fitness: {best_fitness}")
    print(f"Best AE genome: {model.__dict__}")

    return model



import torch
import random
import numpy as np
import torch.nn.functional as F

from torch import nn


class Autoencoder(nn.Module):

    def __init__(self, solution, dataset_shape):
        super(Autoencoder, self).__init__()

        self.encoding_layers = nn.ModuleList()
        self.decoding_layers = nn.ModuleList()
        self.bottleneck_size = 0
        self.shape = self.get_shape(solution[0])
        self.layer_step = self.get_layer_step(solution[1], dataset_shape)
        self.layers = self.get_layers(solution[2], self.layer_step, dataset_shape)
        self.activation = self.get_activation(solution[3])
        self.epochs = self.get_epochs(solution[4])
        self.learning_rate = self.get_learning_rate(solution[5])

        self.generate_autoencoder(self.shape,
                                  self.layers,

                                  dataset_shape,
                                  self.layer_step)

        self.optimizer = self.get_optimizer(solution[6])

        print(
            f"y1: Shape: {self.shape}\n"
            f"y2: Layer step: {self.layer_step}\n"
            f"y3: Layers: {self.layers}\n"
            f"y4: Activation function: {self.activation}\n"
            f"y5: Epochs: {self.epochs}\n"
            f"y6: Learning rate: {self.learning_rate}\n"
            f"y7: Optimizer: {self.optimizer}\n"
            f"Bottleneck size: {self.bottleneck_size}\n"
            f"Encoder: {self.encoding_layers}\n"
            f"Decoder: {self.decoding_layers}\n")

    def forward(self, x):
        """Flipping shape of tensors"""
        x = x.reshape(x.shape[1], x.shape[0])

        for layer in self.encoding_layers:
            x = self.activation(layer(x))

        for layer in self.decoding_layers:
            x = self.activation(layer(x))

        """Flipping back to original shape"""
        x = x.reshape(x.shape[1], x.shape[0])

        return x

    def get_shape(self, gene):
        gene = np.array([gene])
        bins = np.array([0.0, 0.5])
        inds = np.digitize(gene, bins)

        if inds[0] - 1 == 0:
            return "SYMMETRICAL"

        elif inds[0] - 1 == 1:
            return "A-SYMMETRICAL"

        else:
            raise ValueError(f"Value not between boundaries 0.0 and 1.0. Value is: {inds[0] - 1}")

    def get_layer_step(self, gene, dataset_shape):
        gene = np.array([gene])
        bins = []
        value = 1 / dataset_shape[1]
        step = value
        for col in range(0, dataset_shape[1]):
            bins.append(step)
            step += value
        bins[-1] = 1.01
        inds = np.digitize(gene, bins)
        return inds[0]

    def get_layers(self, gene, layer_step, dataset_shape):
        if layer_step == 0:
            max_layers = dataset_shape[1]
            return max_layers

        else:
            max_layers = round(dataset_shape[1] / layer_step)

        if max_layers == 1:
            return 1

        else:
            gene = np.array([gene])

            bins = []
            value = 1 / max_layers
            step = value
            for col in range(0, max_layers):
                bins.append(step)
                step += value
            bins[-1] = 1.01
            inds = np.digitize(gene, bins)

            return inds[0]

    def get_activation(self, gene):
        gene = np.array([gene])
        bins = np.array([0.0, 0.125, 0.25, 0.375, 0.500, 0.625, 0.750, 0.875, 1.01])
        inds = np.digitize(gene, bins)

        if inds[0] - 1 == 0:
            return F.elu

        elif inds[0] - 1 == 1:
            return F.relu

        elif inds[0] - 1 == 2:
            return F.leaky_relu

        elif inds[0] - 1 == 3:
            return F.rrelu

        elif inds[0] - 1 == 4:
            return F.selu

        elif inds[0] - 1 == 5:
            return F.celu

        elif inds[0] - 1 == 6:
            return F.gelu

        elif inds[0] - 1 == 7:
            return torch.tanh

        else:

            raise ValueError(f"Value not between boundaries 0.0 and 1.0. Value is: {inds[0] - 1}")

    def get_epochs(self, gene):
        gene = np.array([gene])
        bins = np.array([0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.60, 0.7, 0.8, 0.9, 1.01])
        inds = np.digitize(gene, bins)

        return inds[0] * 10 + 100

    def get_learning_rate(self, gene):
        gene = np.array([gene])
        bins = []
        value = 1 / 1000
        step = value
        for col in range(0, 1000):
            bins.append(step)
            step += value
        bins[-1] = 1.01
        inds = np.digitize(gene, bins)
        lr = np.array(bins)[inds[0]]

        return lr

    def generate_autoencoder(self, shape, layers, dataset_shape, layer_step):
        if shape == "SYMMETRICAL":

            i = dataset_shape[1]
            z = dataset_shape[1] - layer_step

            while layers != 0:
                """Minimum depth reached"""
                if z < 1:
                    self.encoding_layers.append(nn.Linear(in_features=i, out_features=z + 1))
                    self.decoding_layers.insert(0, nn.Linear(in_features=z + 1, out_features=i))
                    self.bottleneck_size = z + 1
                    break

                self.encoding_layers.append(nn.Linear(in_features=i, out_features=z))
                self.decoding_layers.insert(0, nn.Linear(in_features=z, out_features=i))
                i = i - layer_step
                z = z - layer_step
                layers = layers - 1

            if len(self.encoding_layers) == 0:
                self.bottleneck_size = 0
            else:
                self.bottleneck_size = self.encoding_layers[-1].out_features

        elif shape == "A-SYMMETRICAL":
            i = dataset_shape[1]
            z = dataset_shape[1] - layer_step

            if layers == 1 or layers == 2:
                self.encoding_layers.append(nn.Linear(in_features=i, out_features=z))
                self.decoding_layers.insert(0, nn.Linear(in_features=z, out_features=i))

            if layers >= 3:
                layers_encoder = random.randint(1, layers)
                layers_decoder = layers - layers_encoder

                encoder_counter = layers_encoder
                decoder_counter = layers_decoder

                if layers_decoder == 0:
                    layers_encoder = layers_encoder - 1
                    layers_decoder = 1

                    encoder_counter = layers_encoder
                    decoder_counter = layers_decoder

                while encoder_counter != 0:

                    if z < 1:
                        self.encoding_layers.append(nn.Linear(in_features=i, out_features=z + 1))
                        self.bottleneck_size = z + 1
                        break

                    self.encoding_layers.append(nn.Linear(in_features=i, out_features=z))

                    i = i - layer_step
                    z = z - layer_step
                    encoder_counter = encoder_counter - 1

                while decoder_counter != 0:

                    if layers_decoder == 1:
                        self.decoding_layers.insert(0, nn.Linear(in_features=i, out_features=dataset_shape[1]))
                        break

                    layer_step = int((dataset_shape[1] - i) / decoder_counter)  # Make more complex logic
                    last_i = i
                    i = i + layer_step
                    z = z + layer_step
                    decoder_counter = decoder_counter - 1

                    self.decoding_layers.append(nn.Linear(in_features=last_i, out_features=i))

            if len(self.encoding_layers) == 0:
                self.bottleneck_size = 0
            else:
                self.bottleneck_size = self.encoding_layers[-1].out_features

    def get_optimizer(self, gene):
        gene = np.array([gene])
        bins = np.array([0.0, 0.167, 0.334, 0.50, 0.667, 0.834, 1.01])
        inds = np.digitize(gene, bins)

        """When AE does not have any layers"""
        if len(list(self.parameters())) == 0:
            return None

        if inds[0] - 1 == 0:
            return torch.optim.Adam(self.parameters(), lr=self.learning_rate)

        elif inds[0] - 1 == 1:
            return torch.optim.Adagrad(self.parameters(), lr=self.learning_rate)

        elif inds[0] - 1 == 2:
            return torch.optim.SGD(self.parameters(), lr=self.learning_rate)

        elif inds[0] - 1 == 3:
            return torch.optim.RAdam(self.parameters(), lr=self.learning_rate)

        elif inds[0] - 1 == 4:
            return torch.optim.ASGD(self.parameters(), lr=self.learning_rate)

        elif inds[0] - 1 == 5:
            return torch.optim.Rprop(self.parameters(), lr=self.learning_rate)

        else:
            raise ValueError(f"Value not between boundaries 0.0 and 1.0. Value is: {inds[0] - 1}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nianet']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'niapy>=2.0.1,<3.0.0',
 'numpy>=1.22.3,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'torch>=1.13.1,<2.0.0']

setup_kwargs = {
    'name': 'nianet',
    'version': '1.1.3',
    'description': 'Designing and constructing neural network topologies using nature-inspired algorithms',
    'long_description': '<p align="center"><img src=".github/NiaNetLogo.png" alt="NiaPy" title="NiaNet"/></p>\n\n---\n[![PyPI Version](https://img.shields.io/badge/pypi-v1.0.0-blue)](https://pypi.org/project/nianet/)\n![PyPI - Python Version](https://img.shields.io/badge/python-3.8-blue)\n[![Downloads](https://static.pepy.tech/badge/nianet)](https://pepy.tech/project/nianet)\n[![GitHub license](https://img.shields.io/badge/license-MIT-green)](https://github.com/SasoPavlic/NiaNet/blob/main/LICENSE)\n### Designing and constructing neural network topologies using nature-inspired algorithms\n\n### Description 📝\n\nThe proposed method NiaNet attempts to pick hyperparameters and AE architecture that will result in a successful encoding and decoding (minimal difference between input and output). NiaNet uses the collection of algorithms available in the library [NiaPy](https://github.com/NiaOrg/NiaPy) to navigate efficiently in waste search-space.\n\n### What it can do? 👀\n\n* **Construct novel AE\'s architecture** using nature-inspired algorithms.\n* It can be utilized for **any kind of dataset**, which has **numerical** values.\n\n### Installation ✅\n\nInstalling NiaNet with pip3: \n```sh\npip3 install nianet\n```\n\n### Documentation 📘\n\n**Annals of Computer Science and Information Systems, Volume 30:**\n[NiaNet: A framework for constructing Autoencoder architectures using nature-inspired algorithms](https://www.sasopavlic.com/publication/nianet-a-framework-for-constructing-autoencoder-architectures-using-nat-ure-inspired-algorithms/)\n\n### Examples\n\nUsage examples can be found [here](examples).\n\n### Getting started 🔨\n\n##### Create your own example:\nIn [examples](examples) folder create the Python file based on the existing [evolve_for_diabetes_dataset.py](examples/evolve_for_diabetes_dataset.py).\n\n##### Change dataset:\nChange the dataset import function as follows:\n```python\nfrom sklearn.datasets import load_diabetes\ndataset = load_diabetes()\n```\n\n##### Specify the search space:\n\nSet the boundaries of your search space with [autoencoder.py](nianet/autoencoder.py).\n\nThe following dimensions can be modified:\n* Topology shape (symmetrical, asymmetrical)\n* Size of input, hidden and output layers\n* Number of hidden layers\n* Number of neurons in hidden layers\n* Activation functions\n* Number of epochs\n* Learning rate\n* Optimizer\n\nYou can run the NiaNet script once your setup is complete.\n##### Running NiaNet script:\n\n`python evolve_for_diabetes_dataset.py`\n\n### HELP ⚠️\n\n**saso.pavlic@student.um.si**\n\n## Acknowledgments 🎓\n\n* NiaNet was developed under the supervision\n  of [doc. dr Iztok Fister ml.](http://www.iztok-jr-fister.eu/)\n  at [University of Maribor](https://www.um.si/en/home-page/).\n\n* This code is a fork of [NiaPy](https://github.com/NiaOrg/NiaPy). I am grateful that the authors chose to\n  open-source their work for future use.\n\n# Cite us\nAre you using NiaNet in your project or research? Please cite us!\n### Plain format\n```\nS. Pavlič, I. F. Jr, and S. Karakatič, “NiaNet: A framework for constructing Autoencoder architectures using nature-inspired algorithms,” in Annals of Computer Science and Information Systems, 2022, vol. 30, pp. 109–116. Accessed: Oct. 08, 2022. [Online]. Available: https://annals-csis.org/Volume_30/drp/192.html\n```\n### Bibtex format\n```\n    @article{NiaPyJOSS2018,\n        author  = {Vrban{\\v{c}}i{\\v{c}}, Grega and Brezo{\\v{c}}nik, Lucija\n                  and Mlakar, Uro{\\v{s}} and Fister, Du{\\v{s}}an and {Fister Jr.}, Iztok},\n        title   = {{NiaPy: Python microframework for building nature-inspired algorithms}},\n        journal = {{Journal of Open Source Software}},\n        year    = {2018},\n        volume  = {3},\n        issue   = {23},\n        issn    = {2475-9066},\n        doi     = {10.21105/joss.00613},\n        url     = {https://doi.org/10.21105/joss.00613}\n    }\n```\n### RIS format\n```\nTY  - CONF\nTI  - NiaNet: A framework for constructing Autoencoder architectures using nature-inspired algorithms\nAU  - Pavlič, Sašo\nAU  - Jr, Iztok Fister\nAU  - Karakatič, Sašo\nT2  - Proceedings of the 17th Conference on Computer Science and Intelligence Systems\nC3  - Annals of Computer Science and Information Systems\nDA  - 2022///\nPY  - 2022\nDP  - annals-csis.org\nVL  - 30\nSP  - 109\nEP  - 116\nLA  - en\nSN  - 978-83-962423-9-6\nST  - NiaNet\nUR  - https://annals-csis.org/Volume_30/drp/192.html\nY2  - 2022/10/08/19:08:20\nL1  - https://annals-csis.org/Volume_30/drp/pdf/192.pdf\nL2  - https://annals-csis.org/Volume_30/drp/192.html\n```\n\n\n\n## License\n\nThis package is distributed under the MIT License. This license can be found online at <http://www.opensource.org/licenses/MIT>.\n\n## Disclaimer\n\nThis framework is provided as-is, and there are no guarantees that it fits your purposes or that it is bug-free. Use it at your own risk!\n',
    'author': 'Sašo Pavlič',
    'author_email': 'saso.pavlic@student.um.si',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SasoPavlic/NiaNet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
